/*
	Copyright (C) 2011 Steve Thomas <steve AT tobtu DOT com>

	This file is part of MySQL323 Collider.

	MySQL323 Collider is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	MySQL323 Collider is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with MySQL323 Collider.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <signal.h>
#include <list>
#include <string>
#include <vector>
#include <map>
#include "common.h"
#include "mysql323thread.h"
#include "keyspace.h"
#include "atomic.h"
#include "workqueue.h"

using namespace std;

struct programArgs
{
	char *collisions, *hash, *hashList, *memory, *outFile, *threads, *resume;
	int even, help, noBitmap, odd, prefix, silent, superSilent, verbose;
};

void usage(char *name);
int parseArgs(int argc, char *argv[], programArgs *args);
void initBitmapAndIndex(uint32_t *bitmap, passwordElement **indexedLookupTable, passwordElement *lookupTable, uint64_t tableSize);
inline uint64_t quickSortPartition(passwordElement *row, uint64_t lo, uint64_t hi);
void sort(passwordElement *row, uint64_t lo, uint64_t hi);
void initPrevHash(uint8_t *prevHash);
//uint64_t countPasswordsOfAGivenSum(uint32_t prefixChars, uint32_t prefixMinChar, uint32_t prefixMaxChar, uint32_t minSum, uint32_t maxSum);
uint64_t listPasswordsOfAGivenSum(passwordElement *lookupTable, uint32_t prefixChars, uint32_t prefixMinChar, uint32_t prefixMaxChar, uint32_t minSum, uint32_t maxSum, uint32_t even, uint64_t tableSize);
void catchCtrlC(int sign);
uint32_t initialize(passwordElement ***indexedLookupTableOut, uint32_t **bitmapOut, uint8_t **prevHashOut, uint32_t prefixMinChar, uint32_t prefixMaxChar, uint32_t prefixChars, uint32_t minSum, uint32_t maxSum, uint32_t bitFieldSize, uint32_t verbose, uint64_t memory, uint32_t even);
void formatSpeed(double speed);
int readHash(const char *hash, mysqlHash &out);
char *str2Hex(char *hex, const char *str);

bool stopCracking = false;

const char *speedPrefixes;
const char SI_PREFIXES[]    = " kMGTPEZY"; // kilo, mega, giga, tera, peta, exa, zetta, yotta
const char OTHER_PREFIXES[] = " tMBTQ";    // thousand, million, billion, trillion, quadrillion

void usage(char *name)
{
	printf("\n                        **** MySQL323 Collider v1.1 ****\n\n"
		"Generates a collision for a MySQL323 hash.\n\n"
#ifdef _WIN32
		"\"%s\" [parameters]\n\n"
#else
		"%s [parameters]\n\n"
#endif
		"*** Required parameters ***\n"
		" -h|--hash=<string>\n"
		"    A MySQL323 hash (such as 7fffffff7fffffff)\n\n"
		" -l|--list=<file>\n"
		"    A file with a list MySQL323 hashes\n\n"
		" -m|--memory=<int>\n"
		"    MiB of memory to allocate for main lookup table\n\n"
		" -t|--threads=<int>\n"
		"    Number of threads\n\n\n"
		"*** Optional parameters ***\n"
		" --help\n"
		"    Display this message and exit\n\n"
		" -c|--collisions=<int>\n"
		"    Number of collisions found before exiting (default 1)\n\n"
		" -e|--even\n"
		"    Run only even hashes (Only applies to -l|--list)\n\n"
		" -n|--no-bitmap\n"
		"    Don't create the 31 bits of the hash bitmap\n\n"
		" -o|--output=<file>\n"
		"    A file to output cracked hashes\n\n"
		" -O|--odd\n"
		"    Run only odd hashes (Only applies to -l|--list)\n\n"
		" -p|--prefix\n"
		"    Use t, M, B, T, and Q for thousand, million, billion, trillion, and\n"
		"    quadrillion. Instead of the SI prefixes k, M, G, T, P, E, Z, and Y.\n\n"
		" -r|--resume=<int>\n"
		"    Resume code\n\n"
		" -s|--silent\n"
		"    Silent mode (suppresses output of current speed)\n\n"
		" -S|--super-silent\n"
		"    Super silent mode (outputs cracked hashes or nothing with -o)\n\n"
		" -v|--verbose\n"
		"    Verbose mode\n\n"
		"Copyright (c) 2011 Steve Thomas <steve AT tobtu DOT com>\n"
		"  This is free software: you can redistribute it and/or modify it under the\n"
		"  terms of the GNU General Public License as published by the Free Software\n"
		"  Foundation, either version 3 of the License, or (at your option) any later\n"
		"  version. There is NO WARRANTY, to the extent permitted by law.\n"
			, name);
}

/**
 * return 0 all good
 * return 1 value used
 * return -1 error
 */
int searchArg(string &arg, char *value, map<string,char**> &argValues, map<string,int*> &argSwitches)
{
	map<string,char**>::iterator itValue;
	map<string,int*  >::iterator itSwitch;
	string argName;
	size_t equal;
	int ret = 0;

	equal = arg.find('=');
	argName = arg.substr(0, equal);
	itValue = argValues.find(argName);
	if (itValue != argValues.end())
	{
		if (equal != string::npos)
		{
			*(itValue->second) = value;
		}
		else if (value != NULL)
		{
			*(itValue->second) = value;
			ret = 1;
		}
		else
		{
			fprintf(stderr, "Error: '-%s' requires a value\n", arg.c_str());
			ret = -1;
		}
	}
	else
	{
		itSwitch = argSwitches.find(argName);
		if (itSwitch != argSwitches.end())
		{
			if (equal == string::npos)
			{
				*(itSwitch->second) = 1;
			}
			else
			{
				fprintf(stderr, "Error: '-%s' does not have a value\n", argName.c_str());
				ret = -1;
			}
		}
		else
		{
			fprintf(stderr, "Error: Unknown argument '-%s'\n", argName.c_str());
			ret = -1;
		}
	}
	return ret;
}

int parseArgs(int argc, char *argv[], programArgs *args)
{
	map<string,char**> argValues;
	map<string,int*  > argSwitches;
	string arg;
	char *value;
	size_t equal;
	int a, b, ret = 0, tmp;
	
	args->even        = 0;
	args->help        = 0;
	args->noBitmap    = 0;
	args->odd         = 0;
	args->prefix      = 0;
	args->silent      = 0;
	args->superSilent = 0;
	args->verbose     = 0;
	args->collisions  = NULL;
	args->hash        = NULL;
	args->hashList    = NULL;
	args->memory      = NULL;
	args->outFile     = NULL;
	args->threads     = NULL;
	args->resume      = NULL;
	
	argSwitches["-even"]          = &(args->even);
	argSwitches["e"]             = &(args->even);
	argSwitches["-help"]         = &(args->help);
	argSwitches["-no-bitmap"]    = &(args->noBitmap);
	argSwitches["n"]             = &(args->noBitmap);
	argSwitches["-odd"]          = &(args->odd);
	argSwitches["O"]             = &(args->odd);
	argSwitches["-prefix"]       = &(args->prefix);
	argSwitches["p"]             = &(args->prefix);
	argSwitches["-silent"]       = &(args->silent);
	argSwitches["s"]             = &(args->silent);
	argSwitches["-super-silent"] = &(args->superSilent);
	argSwitches["S"]             = &(args->superSilent);
	argSwitches["-verbose"]      = &(args->verbose);
	argSwitches["v"]             = &(args->verbose);
	argValues["-collisions"]     = &(args->collisions);
	argValues["c"]               = &(args->collisions);
	argValues["-hash"]           = &(args->hash);
	argValues["h"]               = &(args->hash);
	argValues["-list"]           = &(args->hashList);
	argValues["l"]               = &(args->hashList);
	argValues["-memory"]         = &(args->memory);
	argValues["m"]               = &(args->memory);
	argValues["-output"]         = &(args->outFile);
	argValues["o"]               = &(args->outFile);
	argValues["-resume"]         = &(args->resume);
	argValues["r"]               = &(args->resume);
	argValues["-threads"]        = &(args->threads);
	argValues["t"]               = &(args->threads);

	for (a = 1; a < argc; a++)
	{
		if (argv[a][0] == '-')
		{
			value = NULL;
			if (argv[a][1] == '-')
			{
				arg = argv[a] + 1;
				equal = arg.find('=');
				if (equal != string::npos)
				{
					value = argv[a] + (equal + 1);
				}
				else if (a + 1 < argc)
				{
					value = argv[a + 1];
				}
				switch (searchArg(arg, value, argValues, argSwitches))
				{
					case  1: a++; break;
					case -1: ret = 1;
				}
			}
			else
			{
				for (b = 1; argv[a][b] != 0 && argv[a][b] != '='; b++)
				{
					arg = argv[a][b];
					if (argv[a][b + 1] == '=')
					{
						arg = argv[a] + b;
						value = argv[a] + (b + 2);
					}
					else if (argv[a][b + 1] == 0 && a + 1 < argc)
					{
						value = argv[a + 1];
					}
					tmp = searchArg(arg, value, argValues, argSwitches);
					if (tmp == 1)
					{
						a++;
						break;
					}
					else if (tmp != 0)
					{
						ret = 1;
						break;
					}
				}
			}
		}
		else
		{
			fprintf(stderr, "Error: Unknown argument '%s'\n", argv[a]);
			ret = 1;
		}
	}
	return ret;
}

int main(int argc, char *argv[])
{
	uint64_t totalPasswords, skipPasswords = 0, memory = 0;
	double speed, totalTime, prefixPasswords;
	MySQL323Thread *threads;
	mySQL323WorkQueueNode *nodes, *node;
	mySQL323WorkQueueData *data;
	mySQL323ThreadArgs *args;
	passwordElement **indexedLookupTable;
	KeySpace *keySpace, *prefixKeySpace;
	double *times, *passwords, *speeds;
	uint32_t *bitmap, *collisions;
	uint8_t *prevHash;
	char *hash, *hexPw;
	programArgs programArgs;
	FILE *fin, *fout;
	string charSet;
	list<string> subKeySpace;
	list<list<string> > subKeySpaces;
	vector<mysqlHash> hashes;
	Atomic keySpacePos(0);
	WorkQueue toMySQL323, fromMySQL323;
	TIMER_TYPE t1, t2, startTime, endTime;
	mysqlHash curHash;
	uint32_t a, hash1, hash2, hash3, numThreads, verbose, targetSum, defaultCollisions = 1, even;
	int ret;

	if (argc == 1)
	{
		usage(argv[0]);
		return 1;
	}
	ret = parseArgs(argc, argv, &programArgs);
	if (programArgs.help)
	{
		usage(argv[0]);
		return ret;
	}
	if (programArgs.hashList == NULL && programArgs.hash == NULL)
	{
		fprintf(stderr, "Error: Missing argument -h|--hash or -l|--list\n");
		ret = 1;
	}
	else if (programArgs.hashList != NULL && programArgs.hash != NULL)
	{
		fprintf(stderr, "Error: Can't have both -h|--hash and -l|--list\n");
		ret = 1;
	}
	if (programArgs.memory == NULL)
	{
		fprintf(stderr, "Error: Missing argument -m|--memory\n");
		ret = 1;
	}
	if (programArgs.threads == NULL)
	{
		fprintf(stderr, "Error: Missing argument -t|--threads\n");
		ret = 1;
	}
	if (programArgs.silent && programArgs.verbose || programArgs.superSilent && programArgs.verbose)
	{
		fprintf(stderr, "Error: Can't be silent and verbose\n");
		ret = 1;
	}
	if (programArgs.even && programArgs.odd)
	{
		fprintf(stderr, "Error: Can't run even and odd hashes\n");
		ret = 1;
	}
	if (ret)
	{
		usage(argv[0]);
		return ret;
	}

	// Resume
	if (programArgs.resume)
	{
		sscanf(programArgs.resume, "%" PRIu64, &skipPasswords);
		keySpacePos.add(skipPasswords);
	}

	// Number of threads
	sscanf(programArgs.threads, "%u", &numThreads);

	// verbose:
	//  0: none
	//  1: Hash cracked output
	//  2: Exit summary
	//  3: Initialization time
	//  4: Current speed
	//  5: Initialization times
	//
	// -S -o 0
	// -S 1
	// -s 3
	// default 4
	// -v 5
	verbose = 4;
	if (programArgs.verbose)
	{
		verbose = 5;
	}
	else if (programArgs.superSilent)
	{
		verbose = 1;
		if (programArgs.outFile)
		{
			verbose = 0;
		}
	}
	else if (programArgs.silent)
	{
		verbose = 3;
	}

	// Number of collisions
	if (programArgs.collisions != NULL)
	{
		sscanf(programArgs.collisions, "%u", &defaultCollisions);
	}

	// Memory
	if (programArgs.memory != NULL)
	{
		sscanf(programArgs.memory, "%u", &(((uint64u*) &memory)->lo32));
		if (memory > 59664)
		{
			fprintf(stderr, "Error: The maximum amount of memory is 59,664 MiB\n");
			return 1;
		}
		memory *= 1024 * 1024;
#ifndef ARC_64
		if (memory > UINT32_MAX)
		{
			fprintf(stderr, "Error: Can't allocate 4 GiB or more of memory\n");
			return 1;
		}
#endif
	}

	// Read hash
	if (programArgs.hash)
	{
		hash = programArgs.hash;
		if (readHash(hash, curHash) != 0)
		{
			fprintf(stderr, "ERROR: Not a valid hash\n");
			return 1;
		}

		even = !(curHash.hash1 & 1);

		// Reverse hash2
		hash3 = curHash.hash2 - curHash.hash1;
		hash3 = curHash.hash2 - ((hash3 << 8) ^ curHash.hash1);
		hash3 = curHash.hash2 - ((hash3 << 8) ^ curHash.hash1);
		curHash.hash2 = curHash.hash2 - ((hash3 << 8) ^ curHash.hash1);
		hashes.push_back(curHash);
		collisions = new uint32_t[1];
		collisions[0] = defaultCollisions;
	}

	// Hash list
	if (programArgs.hashList)
	{
		fin = fopen(programArgs.hashList, "r");
		if (fin == NULL)
		{
			perror(programArgs.hashList);
			return 1;
		}
		hash = new char [32];
		even = 0x80000000;
		if (programArgs.even)
		{
			even = 1;
		}
		else if (programArgs.odd)
		{
			even = 0;
		}
		while (1)
		{
			if (fgets(hash, 32, fin) == NULL)
			{
				if (feof(fin))
				{
					break;
				}
				perror(programArgs.hashList);
				delete [] hash;
				return 1;
			}

			// Not an empty line
			if (hash[0] != 0 && (hash[0] != '\n' || hash[1] != 0) && (hash[0] != '\r' || hash[1] != '\n' || hash[2] != 0))
			{
				if (readHash(hash, curHash) != 0)
				{
					fprintf(stderr, "ERROR: Not a valid hash \"%s\"\n", hash);
					delete [] hash;
					return 1;
				}

				if (even == 0x80000000)
				{
					even = !(curHash.hash1 & 1);
				}
				if (even == !(curHash.hash1 & 1))
				{
					// Reverse hash2
					hash3 = curHash.hash2 - curHash.hash1;
					hash3 = curHash.hash2 - ((hash3 << 8) ^ curHash.hash1);
					hash3 = curHash.hash2 - ((hash3 << 8) ^ curHash.hash1);
					curHash.hash2 = curHash.hash2 - ((hash3 << 8) ^ curHash.hash1);
					hashes.push_back(curHash);
				}
				else if (programArgs.even == 0 && programArgs.odd == 0)
				{
					fprintf(stderr, "Error: You must use -e|--even or -O|--odd\n");
					return 1;
				}
			}
		}
		delete [] hash;
		fclose(fin);

		if (hashes.size() == 0)
		{
			fprintf(stderr, "Error: No hashes found\n");
			return 1;
		}
		collisions = new uint32_t[hashes.size()];
		for (a = 0; a < hashes.size(); a++)
		{
			collisions[a] = defaultCollisions;
		}
		printf("Loaded %u hashes\n", (uint32_t) hashes.size());
	}

	// Output cracked hashes
	fout = NULL;
	if (programArgs.outFile != NULL)
	{
		fout = fopen(programArgs.outFile, "a");
		if (fout == NULL)
		{
			perror(programArgs.outFile);
			return 1;
		}
	}

	// Init keySpace
	// *
	// **
	// ***
	// ****
	// *****
	// ******
	// *******
//	charSet = "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~";
	charSet = "\"$&(*,.02468:<>@BDFHJLNPRTVXZ\\^`bdfhjlnprtvxz|~";
	for (a = 0; a < 7; a++)
	{
		subKeySpace.push_back(charSet);
		subKeySpaces.push_back(subKeySpace);
	}
	keySpace = new KeySpace(subKeySpaces);
	subKeySpace.clear();
	subKeySpaces.clear();

	// Init prefixKeySpace
	// ******
	charSet = "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~";
	for (a = 0; a < 6; a++)
	{
		subKeySpace.push_back(charSet);
	}
	subKeySpaces.push_back(subKeySpace);
	prefixKeySpace = new KeySpace(subKeySpaces);
	subKeySpace.clear();
	subKeySpaces.clear();

	hexPw = new char[2 * (keySpace->getMaxPwLen() + prefixKeySpace->getMaxPwLen()) + 5];

	// Init
	targetSum = 477;
	if (even)
	{
		targetSum = 478;
	}
	if (programArgs.noBitmap)
	{
		bitmap = NULL;
		if (initialize(&indexedLookupTable, NULL, &prevHash, '!', '~', 6, targetSum, targetSum, 31, verbose, memory, even))
		{
			// Error
			fprintf(stderr, "Error\n");
			delete keySpace;
			delete prefixKeySpace;
			return 1;
		}
	}
	else if (initialize(&indexedLookupTable, &bitmap, &prevHash, '!', '~', 6, targetSum, targetSum, 31, verbose, memory, even))
	{
		// Error
		fprintf(stderr, "Error\n");
		delete keySpace;
		delete prefixKeySpace;
		return 1;
	}
	prefixPasswords = indexedLookupTable[1 << 24] - indexedLookupTable[0];
	prefixPasswords *= 47;

	// Init threads
	if (programArgs.prefix)
	{
		speedPrefixes = OTHER_PREFIXES;
	}
	else
	{
		speedPrefixes = SI_PREFIXES;
	}
	totalPasswords = keySpace->getKeySpace();
	times = new double[numThreads];
	passwords = new double[numThreads];
	speeds = new double[numThreads];
	threads = new MySQL323Thread[numThreads];
	nodes = new mySQL323WorkQueueNode[numThreads * 2];
	data = new mySQL323WorkQueueData[numThreads * 2];
	args = new mySQL323ThreadArgs[numThreads];
	for (a = 0; a < numThreads; a++)
	{
		times[a] = 0;
		passwords[a] = 0;
		speeds[a] = 0;
		args[a].inQueue            = &toMySQL323;
		args[a].outQueue           = &fromMySQL323;
		args[a].keySpacePos        = &keySpacePos;
		args[a].keySpace           = keySpace;
		args[a].prefixKeySpace     = prefixKeySpace;
		args[a].indexedLookupTable = indexedLookupTable;
		args[a].bitmap             = bitmap;
		args[a].prevHash           = prevHash;
		args[a].threadId           = a;
		threads[a].start(args + a);
	}

	if (verbose > 1)
	{
		TIMER_FUNC(startTime);
		t1 = startTime;
		if (verbose > 3)
		{
			printf("\rCalculating speed...");
			fflush(stdout);
		}
	}

	// Initial work
	for (a = 0; a < 2 * numThreads; a++)
	{
		nodes[a].data  = data + a;
		data[a].hashes = hashes;
		data[a].sum    = targetSum + 7;
//		data[a].minSum = targetSum + 7;
//		data[a].maxSum = targetSum + 7;
		toMySQL323.push((workQueueNode*) (nodes + a));
	}

	// Catch Ctrl+C
	signal(SIGINT, &catchCtrlC);

	// Process completed work
	node = (mySQL323WorkQueueNode*) fromMySQL323.pop();
	while (keySpacePos.get() < totalPasswords)
	{
		// Update speed
		if (verbose > 3)
		{
			a = node->data->threadId;
			times[a]     += node->data->time;
			passwords[a] += node->data->count;

			TIMER_FUNC(t2);
			if (TIMER_DIFF(t1, t2) >= 5)
			{
				speed = 0;
				for (a = 0; a < numThreads; a++)
				{
					if (times[a] != 0)
					{
						speeds[a] = prefixPasswords * passwords[a] / times[a];
						speed += speeds[a];
					}
					times[a] = 0;
					passwords[a] = 0;
				}
				printf("\r");
				formatSpeed(speed);
				printf(" [");
				for (a = 0; a < numThreads; a++)
				{
					if (a != 0)
					{
						printf(" ");
					}
					printf("%0.1f%%", speeds[a] / speed * 100);
				}
				printf("] ");
				fflush(stdout);
				t1 = t2;
			}
		}

		// Found passwords
		if (node->data->foundPws.size() != 0)
		{
			list<pair<uint32_t,string> >::iterator it;
			for (it = node->data->foundPws.begin(); node->data->foundPws.end() != it; ++it)
			{
				str2Hex(hexPw, it->second.c_str());
				hash1 = node->data->hashes[it->first].hash1;
				hash2 = node->data->hashes[it->first].hash2;
				if (verbose > 0)
				{
					if (verbose > 3)
					{
						printf("\n");
					}
					printf("%08x%08x:%s:%s\n", hash1, hash2 + ((hash2 << 8) ^ hash1), hexPw, it->second.c_str());
				}
				if (fout != NULL)
				{
					fprintf(fout, "%08x%08x:%s:%s\n", hash1, hash2 + ((hash2 << 8) ^ hash1), hexPw, it->second.c_str());
					fflush(fout);
				}
				a = it->first;
				if (a >= hashes.size())
				{
					a = hashes.size() - 1;
				}
				for (; (int) a >= 0; a--)
				{
					if (hashes[a].hash1 == hash1 && hashes[a].hash2 == hash2)
					{
						if (collisions[a]-- <= 1)
						{
							hashes.erase(hashes.begin() + a);
							for (; a < hashes.size(); a++)
							{
								collisions[a] = collisions[a + 1];
							}
						}
						break;
					}
				}
			}
			node->data->foundPws.clear();
		}

		// Done?
		if (stopCracking || hashes.size() == 0)
		{
			break;
		}

		// Get next node
		if (hashes.size() != node->data->hashes.size())
		{
			node->data->hashes = hashes;
		}
		toMySQL323.push((workQueueNode*) node);
		node = (mySQL323WorkQueueNode*) fromMySQL323.pop();
	}
	for (a = 0; a < numThreads; a++)
	{
		threads[a].stopRunning();
		threads[a].waitForStopRunning();
	}

	// Found passwords
	fromMySQL323.flush();
	node = (mySQL323WorkQueueNode*) fromMySQL323.pop();
	while (node != NULL)
	{
		if (node->data->foundPws.size() != 0)
		{
			list<pair<uint32_t,string> >::iterator it;
			for (it = node->data->foundPws.begin(); node->data->foundPws.end() != it; ++it)
			{
				str2Hex(hexPw, it->second.c_str());
				hash1 = node->data->hashes[it->first].hash1;
				hash2 = node->data->hashes[it->first].hash2;
				if (verbose > 0)
				{
					if (verbose > 3)
					{
						printf("\n");
					}
					printf("%08x%08x:%s:%s\n", hash1, hash2 + ((hash2 << 8) ^ hash1), hexPw, it->second.c_str());
				}
				if (fout != NULL)
				{
					fprintf(fout, "%08x%08x:%s:%s\n", hash1, hash2 + ((hash2 << 8) ^ hash1), hexPw, it->second.c_str());
					fflush(fout);
				}
				a = it->first;
				if (a >= hashes.size())
				{
					a = hashes.size() - 1;
				}
				for (; (int) a >= 0; a--)
				{
					if (hashes[a].hash1 == hash1 && hashes[a].hash2 == hash2)
					{
						if (collisions[a]-- <= 1)
						{
							hashes.erase(hashes.begin() + a);
							for (; a < hashes.size(); a++)
							{
								collisions[a] = collisions[a + 1];
							}
						}
						break;
					}
				}
			}
			node->data->foundPws.clear();
		}
		node = (mySQL323WorkQueueNode*) fromMySQL323.pop();
	}
	if (fout != NULL)
	{
		fclose(fout);
	}

	// Stats
	if (verbose > 1)
	{
		TIMER_FUNC(endTime);
		if (totalPasswords > keySpacePos.get())
		{
			totalPasswords = keySpacePos.get();
		}
		totalPasswords -= skipPasswords;
		totalTime = TIMER_DIFF(startTime, endTime);
		if (verbose > 3)
		{
			printf("\n");
		}
		printf("Crack time:    %0.3f seconds\n"
		       "Average speed: ", totalTime);
		if (totalTime != 0)
		{
			formatSpeed(prefixPasswords * (totalPasswords / totalTime));
		}
		else
		{
			printf("n/a");
		}
		printf("\n");
	}

	// Resume code
	if (stopCracking && hashes.size() != 0)
	{
		printf("Resume code:   %" PRIu64 "\n", keySpacePos.get());
	}

	// Clean up
	delete keySpace;
	delete prefixKeySpace;
	delete [] collisions;
	delete [] hexPw;
	delete [] times;
	delete [] passwords;
	delete [] speeds;
	delete [] threads;
	delete [] nodes;
	delete [] data;
	delete [] args;
	delete [] *indexedLookupTable;
	delete [] indexedLookupTable;
	if (bitmap != NULL)
	{
		delete [] bitmap;
	}
	delete [] prevHash;
	return 0;
}

void catchCtrlC(int sign)
{
	stopCracking = true;
}

uint32_t initialize(
	passwordElement ***indexedLookupTableOut,
	uint32_t           **bitmapOut,
	uint8_t            **prevHashOut,
	uint32_t prefixMinChar, uint32_t prefixMaxChar,
	uint32_t prefixChars,
	uint32_t minSum,        uint32_t maxSum,
	uint32_t bitFieldSize,  uint32_t verbose, uint64_t memory, uint32_t even)
{
	uint64_t tableSize = 0;
	passwordElement *lookupTable;
	TIMER_TYPE startTime, t1, t2;

	// targetSum = (length * (charSetLen - 1)) / 2
	if (prefixMinChar > prefixMaxChar ||
		prefixChars == 0 || minSum > maxSum ||
		bitFieldSize < 30 || bitFieldSize > 62 ||
		prefixMinChar * prefixChars > maxSum ||
		prefixMaxChar * prefixChars < minSum

		|| minSum != maxSum || bitFieldSize != 31)
	{
		return 1;
	}

	if (verbose > 2)
	{
		printf("Initializing...\n");
		TIMER_FUNC(startTime);
	}

	// Generate prevHash
	if (verbose > 4)
	{
		printf("Generating constants...\n");
		t1 = startTime;
	}
	*prevHashOut = new uint8_t[64 * 64 * 32];
	initPrevHash(*prevHashOut);
	if (verbose > 4)
	{
		TIMER_FUNC(t2);
		printf("Took %0.2f seconds\n", TIMER_DIFF(t1, t2));
		t1 = t2;
	}

	// Create lookupTable
	tableSize = memory / sizeof(passwordElement);
	if (verbose > 4)
	{
		printf("Creating lookup table (%"PRIu64" Bytes)...\n", tableSize * sizeof(passwordElement));
	}
#ifdef ARC_64
	lookupTable = new passwordElement[tableSize];
#else
	if (tableSize <= UINT32_MAX)
	{
		lookupTable = new passwordElement[(uint32_t) tableSize];
	}
	else
	{
		delete [] *prevHashOut;
		return 1;
	}
#endif
	if (bitmapOut != NULL)
	{
		*bitmapOut = new uint32_t[((uint32_t) 1 << 31) / 32];
	}
	*indexedLookupTableOut = new passwordElement*[(1 << 24) + 1];
	tableSize = listPasswordsOfAGivenSum(lookupTable, prefixChars, prefixMinChar, prefixMaxChar, minSum, maxSum, even, tableSize);
	if (verbose > 4)
	{
		TIMER_FUNC(t2);
		printf("Took %0.2f seconds\n", TIMER_DIFF(t1, t2));
		t1 = t2;
	}

	// Sort lookupTable
	if (verbose > 4)
	{
		printf("Sorting lookup table...\n");
	}
	sort(lookupTable, 0, tableSize - 1);
	if (verbose > 4)
	{
		TIMER_FUNC(t2);
		printf("Took %0.2f seconds\n", TIMER_DIFF(t1, t2));
		t1 = t2;
	}

	// Create bitmap and index
	if (verbose > 4)
	{
		printf("Creating bitmap table and index...\n");
	}
	if (bitmapOut != NULL)
	{
		initBitmapAndIndex(*bitmapOut, *indexedLookupTableOut, lookupTable, tableSize);
	}
	else
	{
		initBitmapAndIndex(NULL, *indexedLookupTableOut, lookupTable, tableSize);
	}
	if (verbose > 4)
	{
		TIMER_FUNC(t2);
		printf("Took %0.2f sec\n", TIMER_DIFF(t1, t2));
		t1 = t2;
		printf("Set up took %0.2f sec\n", TIMER_DIFF(startTime, t2));
	}
	else if (verbose > 2)
	{
		TIMER_FUNC(t2);
		printf("Took %0.2f sec\n", TIMER_DIFF(startTime, t2));
	}
	return 0;
}

void initBitmapAndIndex(uint32_t *bitmap, passwordElement **indexedLookupTable, passwordElement *lookupTable, uint64_t tableSize)
{
	uint32_t lastMiniIndex = 0, miniIndex;

	indexedLookupTable[0] = lookupTable;
	if (bitmap != NULL)
	{
		for (uint32_t a = 0; a < ((uint32_t) 1 << 31) / 32; a++)
		{
			bitmap[a] = 0;
		}
	}
	for (uint64_t pos = 0; pos < tableSize; pos++)
	{
		miniIndex = lookupTable[pos].hash2 >> 7;
		while (miniIndex > lastMiniIndex)
		{
			lastMiniIndex++;
			indexedLookupTable[lastMiniIndex] = lookupTable + pos;
		}
		if (bitmap != NULL)
		{
			bitmap[lookupTable[pos].hash2 >> 5] |= 1 << (lookupTable[pos].hash2 & 0x1f);
		}
	}
	while ((1 << 24) > lastMiniIndex)
	{
		lastMiniIndex++;
		indexedLookupTable[lastMiniIndex] = lookupTable + tableSize;
	}
}

uint64_t quickSortPartition(passwordElement *row, uint64_t lo, uint64_t hi)
{
	passwordElement tmp;

	tmp = row[lo];
	while (lo < hi)
	{
		while (lo < hi && row[hi].hash >= tmp.hash)
		{
			hi--;
		}
		row[lo] = row[hi];
		while (lo < hi && row[lo].hash <= tmp.hash)
		{
			lo++;
		}
		row[hi] = row[lo];
	}
	row[lo] = tmp;
	return lo;
}

void sort(passwordElement *row, uint64_t lo, uint64_t hi)
{
	uint64_t pivot;

	pivot = quickSortPartition(row, lo, hi);
	if (lo + 1 < pivot)
	{
		sort(row, lo, pivot - 1);
	}
	if (pivot + 1 < hi)
	{
		sort(row, pivot + 1, hi);
	}
}
/*
uint64_t countPasswordsOfAGivenSum(uint32_t prefixChars, uint32_t prefixMinChar, uint32_t prefixMaxChar, uint32_t minSum, uint32_t maxSum)
{
	uint64_t count = 0;
	uint32_t minTargetSums[MAX_LEN], maxTargetSums[MAX_LEN], chars[MAX_LEN];
	uint32_t pos, sum, ch;

	// Init minTargetSums and maxTargetSums
	for (pos = 0; pos < prefixChars; pos++)
	{
		chars[pos] = prefixMinChar;
		minTargetSums[pos] = 0;
		if (minSum > (prefixChars - pos - 1) * prefixMaxChar)
		{
			minTargetSums[pos] = minSum - (prefixChars - pos - 1) * prefixMaxChar;
		}
		maxTargetSums[pos] = maxSum - (prefixChars - pos - 1) * prefixMinChar;
	}

	pos = 0;
	sum = prefixMinChar;
	prefixChars--;
	if (chars[0] < minTargetSums[0])
	{
		sum = minTargetSums[0];
		chars[0] = minTargetSums[0];
	}
	while (1)
	{
		if (chars[pos] <= prefixMaxChar)
		{
			if (pos < prefixChars)
			{
				if (sum <= maxTargetSums[pos] && sum >= minTargetSums[pos])
				{
					pos++;
					ch = prefixMinChar;
					if (sum + prefixMinChar < minTargetSums[pos])
					{
						ch = minTargetSums[pos] - sum;
					}
					sum += ch;
					chars[pos] = ch;
				}
				else
				{
					if (pos == 0)
						break;
					sum -= chars[pos] - 1;
					pos--;
					chars[pos]++;
				}
			}
			else
			{
				if (sum <= maxTargetSums[pos] && sum >= minTargetSums[pos])
				{
					count++;
				}
				sum -= chars[pos] - 1;
				pos--;
				chars[pos]++;
			}
		}
		else
		{
			if (pos == 0)
				break;
			sum -= chars[pos] - 1;
			pos--;
			chars[pos]++;
		}
	}
	return count;
}
*/
uint64_t listPasswordsOfAGivenSum(passwordElement *lookupTable, uint32_t prefixChars, uint32_t prefixMinChar, uint32_t prefixMaxChar, uint32_t minSum, uint32_t maxSum, uint32_t even, uint64_t tableSize)
{
	uint64_t tablePos, pwNum[MAX_LEN] = {0}, pwPow[MAX_LEN];
	uint32_t chars[MAX_LEN], minTargetSums[MAX_LEN], maxTargetSums[MAX_LEN];
	uint32_t state1Arr[MAX_LEN], state1PartArr[MAX_LEN], state2Arr[MAX_LEN], stepArr[MAX_LEN];
	uint32_t pos, sum, charSetLen, state1, state2, ch;
	uint32_t evenOdd = 1;

	if (even)
	{
		evenOdd = 0;
	}

	// Init pwPow
	pwPow[prefixChars - 1] = 1;
	charSetLen = prefixMaxChar - prefixMinChar + 1;
	for (pos = prefixChars - 2; pos > 0; pos--)
	{
		pwPow[pos] = pwPow[pos + 1] * charSetLen;
	}
	pwPow[0] = pwPow[1] * charSetLen;

	// Init minTargetSums and maxTargetSums
	for (pos = 0; pos < prefixChars; pos++)
	{
		minTargetSums[pos] = 0;
		if (minSum > (prefixChars - pos - 1) * prefixMaxChar)
		{
			minTargetSums[pos] = minSum - (prefixChars - pos - 1) * prefixMaxChar + 7;
		}
		maxTargetSums[pos] = maxSum - (prefixChars - pos - 1) * prefixMinChar + 7;
	}

	pos = 0;
	tablePos = 0;
	prefixChars--;
	sum              = 7 + prefixMinChar;
	state1Arr[0]     = 0x50305735;
	state2Arr[0]     = 0x12345671;
	stepArr[0]       = 60;
	state1PartArr[0] = 60 * prefixMinChar + 0x30573500;
	state1           = 0x50305735 ^ state1PartArr[0];
	state2           = 0x12345671 + (0x34567100 ^ state1);
	chars[0]         = prefixMinChar;
	while (1)
	{
		if (chars[pos] <= prefixMaxChar)
		{
			if (pos < prefixChars)
			{
				if (sum <= maxTargetSums[pos] && sum >= minTargetSums[pos])
				{
					pos++;
					ch = prefixMinChar;
					pwNum[pos] = pwNum[pos - 1];
					if (sum + prefixMinChar < minTargetSums[pos])
					{
						ch = minTargetSums[pos] - sum;
						pwNum[pos] = pwNum[pos - 1] + (ch - prefixMinChar) * pwPow[pos];
					}
					state1Arr[pos]      = state1;
					state2Arr[pos]      = state2;
					stepArr[pos]        = (state1 & 63) + sum;
					state1PartArr[pos]  = stepArr[pos] * ch + (state1 << 8);
					state1              = state1 ^ state1PartArr[pos];
					state2              = state2 + ((state2 << 8) ^ state1);
					sum                += ch;
					chars[pos]          = ch;
				}
				else
				{
					if (pos == 0)
						break;
					sum -= chars[pos] - 1;
					pos--;
					pwNum[pos]         += pwPow[pos];
					state1PartArr[pos] += stepArr[pos];
					state1              = state1Arr[pos] ^ state1PartArr[pos];
					state2              = state2Arr[pos] + ((state2Arr[pos] << 8) ^ state1);
					chars[pos]++;
				}
			}
			else
			{
				if (sum <= maxTargetSums[pos] && sum >= minTargetSums[pos])
				{
					if ((state1 & 1) == evenOdd)
					{
/*
						lookupTable[tablePos].hash1 = state1 & 0x7fffffff;
						lookupTable[tablePos].hash2 = state2 & 0x7fffffff;
						lookupTable[tablePos].pw    = pwNum[pos];
						tablePos++;
						if (tablePos >= tableSize)
						{
							break;
						}
*/
//						lookupTable[tablePos].hash1 = state1 & 0x7fffffff;
//						lookupTable[tablePos].hash2 = state2 & 0x7fffffff;
//						lookupTable[tablePos].pw    = pwNum[pos];
						tablePos++;
						if (tablePos >= tableSize*2)
						{
							printf("%" PRIu64 "\n", pwNum[pos]);
							getchar();
							exit(0);
							break;
						}
					}
				}
				sum -= chars[pos] - 1;
				pos--;
				pwNum[pos]         += pwPow[pos];
				state1PartArr[pos] += stepArr[pos];
				state1              = state1Arr[pos] ^ state1PartArr[pos];
				state2              = state2Arr[pos] + ((state2Arr[pos] << 8) ^ state1);
				chars[pos]++;
			}
		}
		else
		{
			if (pos == 0)
				break;
			sum -= chars[pos] - 1;
			pos--;
			pwNum[pos]         += pwPow[pos];
			state1PartArr[pos] += stepArr[pos];
			state1              = state1Arr[pos] ^ state1PartArr[pos];
			state2              = state2Arr[pos] + ((state2Arr[pos] << 8) ^ state1);
			chars[pos]++;
		}
	}
	return tablePos;
}

void initPrevHash(uint8_t *prevHash)
{
	uint32_t revHash, sum, ch, hash;

	for (revHash = 0; revHash < 64; revHash++)
	{
		for (sum = 0; sum < 64; sum++)
		{
			for (ch = 0; ch < 32; ch++)
			{
				hash = revHash ^ (((sum + revHash) * ch * 2) & 63);
				prevHash[(hash << 11) | (sum << 5) | ch] = revHash;
			}
		}
	}
}

void formatSpeed(double speed)
{
	uint32_t a;
	char prefix;

	for (a = 0; speedPrefixes[a + 1] != 0; a++)
	{
		if (speed < 1000)
		{
			break;
		}
		speed /= 1000;
	}
	prefix = speedPrefixes[a];
	if (speed < 10)
	{
		printf("%0.3f %cp/s", speed, prefix);
	}
	else if (speed < 100)
	{
		printf("%0.2f %cp/s", speed, prefix);
	}
	else if (speed < 1000)
	{
		printf("%0.1f %cp/s", speed, prefix);
	}
	else
	{
		printf(">999 %cp/s", prefix);
	}
}

int readHash(const char *hash, mysqlHash &out)
{
	uint32_t a, h1 = 0, h2 = 0;

	out.hash1 = 0;
	out.hash2 = 0;
	for (a = 0; a < 8; a++)
	{
		if (hash[a] >= '0' && hash[a] <= '9')
		{
			h1 <<= 4;
			h1 |= hash[a] - '0';
		}
		else if (hash[a] >= 'A' && hash[a] <= 'F')
		{
			h1 <<= 4;
			h1 |= hash[a] - 'A' + 10;
		}
		else if (hash[a] >= 'a' && hash[a] <= 'f')
		{
			h1 <<= 4;
			h1 |= hash[a] - 'a' + 10;
		}
		else
		{
			return 1;
		}
	}
	for (; a < 16; a++)
	{
		if (hash[a] >= '0' && hash[a] <= '9')
		{
			h2 <<= 4;
			h2 |= hash[a] - '0';
		}
		else if (hash[a] >= 'A' && hash[a] <= 'F')
		{
			h2 <<= 4;
			h2 |= hash[a] - 'A' + 10;
		}
		else if (hash[a] >= 'a' && hash[a] <= 'f')
		{
			h2 <<= 4;
			h2 |= hash[a] - 'a' + 10;
		}
		else
		{
			return 1;
		}
	}
	if (h1 > 0x7fffffff || h2 > 0x7fffffff || (hash[16] != 0 && (hash[16] != '\n' || hash[17] != 0) && (hash[16] != '\r' || hash[17] != '\n' || hash[18] != 0)))
	{
		return 1;
	}
	out.hash1 = h1;
	out.hash2 = h2;
	return 0;
}

char *str2Hex(char *hex, const char *str)
{
	uint32_t a;
	static const char HEX[] = "0123456789abcdef";

	for (a = 0; *str; a += 2, str++)
	{
		hex[a]     = HEX[((unsigned char) *str) >> 4];
		hex[a + 1] = HEX[((unsigned char) *str) & 0xf];
	}
	hex[a] = 0;
	return hex;
}
